#include "modelcontroller.h"
#include "skingeometry.h"
#include <QQuick3DTextureData>

ModelController::ModelController(QQmlApplicationEngine& engine, QObject* parent)
    : QObject(parent)
    , m_rootNode(nullptr)
    , m_engine(engine)
    , m_geometry(nullptr)
    , m_texture(nullptr)
    , m_material(nullptr)
    , m_textureData(nullptr)
{

}

void ModelController::setRootNode(QQuick3DObject* rootNode)
{
    m_rootNode = rootNode;
}

void ModelController::loadModel()
{
    qDebug() << "ModelController::loadModel";
    QQmlComponent component(&m_engine, QUrl("qrc:/handlers/PrimitiveHandler.qml"));
    if(!component.isReady())
        return;

    if(m_geometry)
    {
        m_geometry->deleteLater();
        m_geometry = nullptr;

        m_texture->deleteLater();
        m_textureData->deleteLater();
        m_material->deleteLater();
        return;
    }

    QQuick3DObject* quickObject = static_cast<QQuick3DObject*>(component.beginCreate(m_engine.rootContext()));
    component.completeCreate();

    m_geometry = new SkinGeometry(quickObject);

    QList<QVector3D> positions = {
        QVector3D(0.0, 0.0, 0.0), // vertex 0
        QVector3D(1.0, 0.0, 0.0), // vertex 1
        QVector3D(0.0, 0.5, 0.0), // vertex 2
        QVector3D(1.0, 0.5, 0.0), // vertex 3
        QVector3D(0.0, 1.0, 0.0), // vertex 4
        QVector3D(1.0, 1.0, 0.0), // vertex 5
        QVector3D(0.0, 1.5, 0.0), // vertex 6
        QVector3D(1.0, 1.5, 0.0), // vertex 7
        QVector3D(0.0, 2.0, 0.0), // vertex 8
        QVector3D(1.0, 2.0, 0.0)  // vertex 9
    };

    m_geometry->setPositions(positions);

    QList<QVector3D> normals = {
        QVector3D(0.0, 0.0, 1.0), // vertex 0
        QVector3D(0.0, 0.0, 1.0), // vertex 1
        QVector3D(0.0, 0.0, 1.0), // vertex 2
        QVector3D(0.0, 0.0, 1.0), // vertex 3
        QVector3D(0.0, 0.0, 1.0), // vertex 4
        QVector3D(0.0, 0.0, 1.0), // vertex 5
        QVector3D(0.0, 0.0, 1.0), // vertex 6
        QVector3D(0.0, 0.0, 1.0), // vertex 7
        QVector3D(0.0, 0.0, 1.0), // vertex 8
        QVector3D(0.0, 0.0, 1.0)  // vertex 9
    };

    m_geometry->setNormals(normals);

    QList<quint32> indexes = {
        0, 1, 3, // triangle 0
        0, 3, 2, // triangle 1
        2, 3, 5, // triangle 2
        2, 5, 4, // triangle 3
        4, 5, 7, // triangle 4
        4, 7, 6, // triangle 5
        6, 7, 9, // triangle 6
        6, 9, 8  // triangle 7
    };

    m_geometry->setIndexes(indexes);

    QList<QVector2D> texcoords = {
        QVector2D(0.0, 0.0), // vertex 0
        QVector2D(1.0, 0.0), // vertex 1
        QVector2D(0.0, 0.25), // vertex 2
        QVector2D(1.0, 0.25), // vertex 3
        QVector2D(0.0, 0.5), // vertex 4
        QVector2D(1.0, 0.5), // vertex 5
        QVector2D(0.0, 0.75), // vertex 6
        QVector2D(1.0, 0.75), // vertex 7
        QVector2D(0.0, 1.0), // vertex 8
        QVector2D(1.0, 1.0)  // vertex 9
    };

    m_geometry->setTexCoords(texcoords);

    quickObject->setProperty("geometry", QVariant::fromValue(m_geometry));

    quickObject->setParent(this);
    quickObject->setParentItem(m_rootNode);


    QQmlComponent textureComponent( &m_engine, QUrl("qrc:/handlers/TextureHandler.qml"));
    if(!textureComponent.isReady())
        return;

    m_texture = static_cast<QQuick3DObject*>(textureComponent.beginCreate(m_engine.rootContext()));

    m_textureData = new QQuick3DTextureData(m_rootNode);
    m_textureData->setSize({256, 256});
    char imageData[256 * 256];
    for(int i = 0; i != 256 * 256; ++i)
    {
        imageData[i] = i / 256;
    }
    QByteArray byteArray;
    byteArray.setRawData(imageData, sizeof(imageData));
    m_textureData->setTextureData(byteArray);
    m_textureData->setFormat(QQuick3DTextureData::Format::R8);
    m_textureData->setHasTransparency(false);

    m_texture->setProperty("textureData", QVariant::fromValue<QObject*>(m_textureData));
    m_texture->setParent(this);
    m_texture->setParentItem(m_rootNode);


    quickObject->setProperty("texture", QVariant::fromValue(m_texture));

    textureComponent.completeCreate();


    QQmlComponent materialComponent(&m_engine, QUrl( "qrc:/handlers/DefaultMaterialHandler.qml"));
    if(!materialComponent.isReady())
        return;

    m_material = static_cast<QQuick3DObject*>(materialComponent.beginCreate(m_engine.rootContext()));

    m_material->setParent(this);
    m_material->setParentItem(m_rootNode);
    m_material->setProperty("diffuseMap", QVariant::fromValue(m_texture));

    quickObject->setProperty("material", QVariant::fromValue(m_material));

    materialComponent.completeCreate();
}
